@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2018-2020 SignalWire, Inc
+ * Copyright (c) 2018-2021 SignalWire, Inc
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -235,7 +235,7 @@ static int ws_client_handshake(kws_t *kws)
 	ks_ssize_t bytes;
 
 	do {
-		bytes = kws_raw_read(kws, kws->buffer + kws->datalen, kws->buflen - kws->datalen, WS_BLOCK);
+		bytes = kws_string_read(kws, kws->buffer + kws->datalen, kws->buflen - kws->datalen, WS_BLOCK);
 	} while (bytes > 0 && !strstr((char *)kws->buffer, "\r\n\r\n"));
 
 	if (bytes > 0) {
@@ -281,7 +281,7 @@ static int ws_server_handshake(kws_t *kws)
 		return -3;
 	}
 
-	while((bytes = kws_raw_read(kws, kws->buffer + kws->datalen, kws->buflen - kws->datalen, WS_BLOCK)) > 0) {
+	while((bytes = kws_string_read(kws, kws->buffer + kws->datalen, kws->buflen - kws->datalen, WS_BLOCK)) > 0) {
 		kws->datalen += bytes;
 		if (strstr(kws->buffer, "\r\n\r\n") || strstr(kws->buffer, "\n\n")) {
 			break;
@@ -444,7 +444,7 @@ KS_DECLARE(ks_ssize_t) kws_raw_read(kws_t *kws, void *data, ks_size_t bytes, int
 		r = -1;
 	}
 
-	if (r > 0) {
+	if (r > 0 && r < bytes) {
 		*((char *)data + r) = '\0';
 	}
 
@@ -455,6 +455,61 @@ KS_DECLARE(ks_ssize_t) kws_raw_read(kws_t *kws, void *data, ks_size_t bytes, int
 	return r;
 }
 
+/*
+ * Blocking read until bytes have been received, failure, or too many retries.
+ */
+static ks_ssize_t kws_raw_read_blocking(kws_t *kws, char *data, ks_size_t max_bytes, int max_retries)
+{
+	ks_ssize_t total_bytes_read = 0;
+	int zero_reads = 0;
+	while (total_bytes_read < max_bytes && zero_reads < max_retries)
+	{
+		int bytes_read = kws_raw_read(kws, data + total_bytes_read, max_bytes - total_bytes_read, WS_BLOCK);
+		if (bytes_read == 0)
+		{
+			zero_reads++;
+			continue;
+		}
+		else if (bytes_read < 0)
+		{
+			break;
+		}
+		total_bytes_read += (ks_ssize_t)bytes_read;
+		zero_reads = 0;
+	}
+	return total_bytes_read;
+}
+
+/**
+ * Read from websocket and store as NULL terminated string. Up to buffer_size - 1 bytes will be read from websocket. Contents of str_buffer is NULL terminated if >0 bytes are read from the websocket.
+ */
+KS_DECLARE(ks_ssize_t) kws_string_read(kws_t *kws, char *str_buffer, ks_size_t buffer_size, int block)
+{
+	if (buffer_size < 1) {
+		return -1;
+	}
+	str_buffer[buffer_size - 1] = '\0';
+	if (buffer_size < 2) {
+		return 0;
+	}
+	return kws_raw_read(kws, str_buffer, buffer_size - 1, block);
+}
+
+/*
+ * Blocking read as from websocket and store as NULL terminated string until buffer_size - 1 have been received, failure, or too many retries. Contents of str_buffer is NULL terminated if >0 bytes are read from the websocket.
+ */
+static ks_ssize_t kws_string_read_blocking(kws_t *kws, char *str_buffer, ks_size_t buffer_size, int max_retries)
+{
+	if (buffer_size < 1) {
+		return -1;
+	}
+	str_buffer[buffer_size - 1] = '\0';
+	if (buffer_size < 2) {
+		return 0;
+	}
+	return kws_raw_read_blocking(kws, str_buffer, buffer_size - 1, max_retries);
+}
+
 KS_DECLARE(ks_ssize_t) kws_raw_write(kws_t *kws, void *data, ks_size_t bytes)
 {
 	int r;
@@ -1097,8 +1152,8 @@ KS_DECLARE(ks_ssize_t) kws_read_frame(kws_t *kws, kws_opcode_t *oc, uint8_t **da
 		return kws_close(kws, WS_NONE);
 	}
 
-	if ((kws->datalen = kws_raw_read(kws, kws->buffer, 9, kws->block)) < 0) {
-		ks_log(KS_LOG_ERROR, "Read frame error because kws_raw_read returned %ld\n", kws->datalen);
+	if ((kws->datalen = kws_string_read(kws, kws->buffer, 9 + 1, kws->block)) < 0) { // read 9 bytes into NULL terminated 10 byte buffer
+		ks_log(KS_LOG_ERROR, "Read frame error because kws_string_read returned %ld\n", kws->datalen);
 		if (kws->datalen == -2) {
 			return -2;
 		}
@@ -1106,11 +1161,11 @@ KS_DECLARE(ks_ssize_t) kws_read_frame(kws_t *kws, kws_opcode_t *oc, uint8_t **da
 	}
 
 	if (kws->datalen < need) {
-		ssize_t bytes = kws_raw_read(kws, kws->buffer + kws->datalen, 9 - kws->datalen, WS_BLOCK);
+		ssize_t bytes = kws_string_read(kws, kws->buffer + kws->datalen, 9 - kws->datalen, WS_BLOCK);
 
 		if (bytes < 0 || (kws->datalen += bytes) < need) {
 			/* too small - protocol err */
-			ks_log(KS_LOG_ERROR, "Read frame error because kws_raw_read: bytes = %ld, datalen = %ld, needed = %ld\n", bytes, kws->datalen, need);
+			ks_log(KS_LOG_ERROR, "Read frame error because kws_string_read: bytes = %ld, datalen = %ld, needed = %ld\n", bytes, kws->datalen, need);
 			return kws_close(kws, WS_NONE);
 		}
 	}
@@ -1146,10 +1201,13 @@ KS_DECLARE(ks_ssize_t) kws_read_frame(kws_t *kws, kws_opcode_t *oc, uint8_t **da
 				need += 4;
 
 				if (need > kws->datalen) {
-					/* too small - protocol err */
-					ks_log(KS_LOG_ERROR, "Read frame error because not enough data for mask\n");
-					*oc = WSOC_CLOSE;
-					return kws_close(kws, WS_NONE);
+					ks_ssize_t bytes = kws_string_read_blocking(kws, kws->buffer + kws->datalen, need - kws->datalen + 1, 10);
+					if (bytes < 0 || (kws->datalen += bytes) < need) {
+						/* too small - protocol err */
+						ks_log(KS_LOG_ERROR, "Read frame error because not enough data for mask\n");
+						*oc = WSOC_CLOSE;
+						return kws_close(kws, WS_NONE);
+					}
 				}
 			}
 
@@ -1163,21 +1221,13 @@ KS_DECLARE(ks_ssize_t) kws_read_frame(kws_t *kws, kws_opcode_t *oc, uint8_t **da
 				need += 8;
 
 				if (need > kws->datalen) {
-					/* too small - protocol err */
-					//*oc = WSOC_CLOSE;
-					//return kws_close(kws, WS_PROTO_ERR);
-
-					more = kws_raw_read(kws, kws->buffer + kws->datalen, (int)(need - kws->datalen), WS_BLOCK);
-
-					if (more < 0 || more < need - kws->datalen) {
-						ks_log(KS_LOG_ERROR, "Read frame error because kws_raw_read: more = %ld, need = %ld, datalen = %ld\n", more, need, kws->datalen);
+					ks_ssize_t bytes = kws_string_read_blocking(kws, kws->buffer + kws->datalen, need - kws->datalen + 1, 10);
+					if (bytes < 0 || (kws->datalen += bytes) < need) {
+						/* too small - protocol err */
+						ks_log(KS_LOG_ERROR, "Read frame error because kws_string_read: more = %ld, need = %ld, datalen = %ld\n", more, need, kws->datalen);
 						*oc = WSOC_CLOSE;
 						return kws_close(kws, WS_NONE);
-					} else {
-						kws->datalen += more;
 					}
-
-
 				}
 
 				u64 = (uint64_t *) kws->payload;
@@ -1189,10 +1239,13 @@ KS_DECLARE(ks_ssize_t) kws_read_frame(kws_t *kws, kws_opcode_t *oc, uint8_t **da
 				need += 2;
 
 				if (need > kws->datalen) {
-					/* too small - protocol err */
-					ks_log(KS_LOG_ERROR, "Read frame error because kws_raw_read: not enough data for packet length\n");
-					*oc = WSOC_CLOSE;
-					return kws_close(kws, WS_NONE);
+					ks_ssize_t bytes = kws_string_read_blocking(kws, kws->buffer + kws->datalen, need - kws->datalen + 1, 10);
+					if (bytes < 0 || (kws->datalen += bytes) < need) {
+						/* too small - protocol err */
+						ks_log(KS_LOG_ERROR, "Read frame error because kws_string_read: not enough data for packet length\n");
+						*oc = WSOC_CLOSE;
+						return kws_close(kws, WS_NONE);
+					}
 				}
 
 				u16 = (uint16_t *) kws->payload;
@@ -1247,7 +1300,7 @@ KS_DECLARE(ks_ssize_t) kws_read_frame(kws_t *kws, kws_opcode_t *oc, uint8_t **da
 
 			while(need) {
 				ks_assert((kws->body + need + kws->rplen) <= (kws->bbuffer + kws->bbuflen));
-				ks_ssize_t r = kws_raw_read(kws, kws->body + kws->rplen, need, WS_BLOCK);
+				ks_ssize_t r = kws_string_read(kws, kws->body + kws->rplen, need + 1, WS_BLOCK);
 
 				if (r < 1) {
 					/* invalid read - protocol err .. */
@@ -1869,8 +1922,11 @@ KS_DECLARE(void) kws_request_reset(kws_request_t *request)
 
 KS_DECLARE(ks_ssize_t) kws_read_buffer(kws_t *kws, uint8_t **data, ks_size_t bytes, int block)
 {
+	if (bytes > kws->buflen) {
+		bytes = kws->buflen;
+	}
 	*data = kws->buffer;
-	return kws_raw_read(kws, kws->buffer, bytes, block);
+	return kws_string_read(kws, kws->buffer, bytes, block);
 }
 
 KS_DECLARE(ks_status_t) kws_keepalive(kws_t *kws)
@@ -1878,7 +1934,7 @@ KS_DECLARE(ks_status_t) kws_keepalive(kws_t *kws)
 	ks_ssize_t bytes = 0;;
 	kws->datalen = 0;
 
-	while ((bytes = kws_raw_read(kws, kws->buffer + kws->datalen, kws->buflen - kws->datalen, WS_BLOCK)) > 0) {
+	while ((bytes = kws_string_read(kws, kws->buffer + kws->datalen, kws->buflen - kws->datalen, WS_BLOCK)) > 0) {
 		kws->datalen += bytes;
 		if (strstr(kws->buffer, "\r\n\r\n") || strstr(kws->buffer, "\n\n")) {
 			return KS_STATUS_SUCCESS;
